(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (global = global || self, factory(global.Rax = {}));
}(this, function (exports) { 'use strict';

  /*
   * Stateful things in runtime
   */
  var Host = {
    mountID: 1,
    component: null,
    isRendering: false,
    dirtyComponents: [],
    // Roots
    rootComponents: {},
    rootInstances: {},
    // Inject
    hook: null,
    driver: null,
    monitor: null
  };

  var RESERVED_PROPS = {
    key: true,
    ref: true
  };

  function traverseChildren(children, result) {
    if (Array.isArray(children)) {
      for (var i = 0, l = children.length; i < l; i++) {
        traverseChildren(children[i], result);
      }
    } else {
      result.push(children);
    }
  }

  function flattenChildren(children) {
    if (children == null) {
      return children;
    }

    var result = [];
    traverseChildren(children, result);

    if (result.length === 1) {
      result = result[0];
    }

    return result;
  }
  function getRenderErrorInfo() {
    if (Host.component) {
      var name = Host.component.getName();

      if (name) {
        return ' Check the render method of `' + name + '`.';
      }
    }

    return '';
  }
  function Element(type, key, ref, props, owner) {
    return {
      // Built-in properties that belong on the element
      type: type,
      key: key,
      ref: ref,
      props: props,
      // Record the component responsible for creating this element.
      _owner: owner
    };
  }
  function createElement(type, config, children) {
    if (type == null) {
      throw Error('createElement: type should not be null or undefined.' + getRenderErrorInfo());
    } // Reserved names are extracted


    var props = {};
    var propName;
    var key = null;
    var ref = null;

    if (config != null) {
      ref = config.ref === undefined ? null : config.ref;
      key = config.key === undefined ? null : String(config.key); // Remaining properties are added to a new props object

      for (propName in config) {
        if (!RESERVED_PROPS[propName]) {
          props[propName] = config[propName];
        }
      }
    }

    var childrenLength = arguments.length - 2;

    if (childrenLength > 0) {
      if (childrenLength === 1 && !Array.isArray(children)) {
        props.children = children;
      } else {
        var childArray = children;

        if (childrenLength > 1) {
          childArray = new Array(childrenLength);

          for (var i = 0; i < childrenLength; i++) {
            childArray[i] = arguments[i + 2];
          }
        }

        props.children = flattenChildren(childArray);
      }
    } // Resolve default props


    if (type && type.defaultProps) {
      var defaultProps = type.defaultProps;

      for (propName in defaultProps) {
        if (props[propName] === undefined) {
          props[propName] = defaultProps[propName];
        }
      }
    }

    return new Element(type, key, ref, props, Host.component);
  }

  function instantiateComponent(element) {
    var instance;

    if (element === undefined || element === null || element === false || element === true) {
      instance = new Host.EmptyComponent();
    } else if (Array.isArray(element)) {
      instance = new Host.FragmentComponent(element);
    } else if (typeof element === 'object' && element.type) {
      // Special case string values
      if (typeof element.type === 'string') {
        instance = new Host.NativeComponent(element);
      } else {
        instance = new Host.CompositeComponent(element);
      }
    } else if (typeof element === 'string' || typeof element === 'number') {
      instance = new Host.TextComponent(element);
    } else {
      throw new Error("Invalid element type: " + element + ". (current: " + (typeof element === 'object' && Object.keys(element) || typeof element) + ")");
    }

    instance._mountIndex = 0;
    return instance;
  }

  function _inheritsLoose(subClass, superClass) {
    subClass.prototype = Object.create(superClass.prototype);
    subClass.prototype.constructor = subClass;
    subClass.__proto__ = superClass;
  }

  /**
   * Base component class.
   */
  var Component =
  /*#__PURE__*/
  function () {
    function Component(props, context, updater) {
      this.props = props;
      this.context = context;
      this.refs = {};
      this.updater = updater;
    }

    var _proto = Component.prototype;

    _proto.isComponentClass = function isComponentClass() {};

    _proto.setState = function setState(partialState, callback) {
      this.updater.setState(this, partialState, callback);
    };

    _proto.forceUpdate = function forceUpdate(callback) {
      this.updater.forceUpdate(this, callback);
    };

    return Component;
  }();

  var rootCounter = 1;

  var Root =
  /*#__PURE__*/
  function (_Component) {
    _inheritsLoose(Root, _Component);

    function Root() {
      var _this;

      _this = _Component.call(this) || this;
      _this.state = {
        // Using fragment instead of null for avoid create a comment node when init mount
        element: []
      };
      _this.rootID = rootCounter++;
      return _this;
    }

    var _proto = Root.prototype;

    _proto.getPublicInstance = function getPublicInstance() {
      return this.getRenderedComponent().getPublicInstance();
    };

    _proto.getRenderedComponent = function getRenderedComponent() {
      return this._internal._renderedComponent;
    };

    _proto.update = function update(element) {
      this.setState({
        element: element
      });
    };

    _proto.render = function render() {
      return this.state.element;
    };

    return Root;
  }(Component);

  /**
   * Instance manager
   */

  var KEY = '__r';
  var instance = {
    set: function set(node, instance) {
      if (!node[KEY]) {
        node[KEY] = instance; // Record root instance to roots map

        if (instance.rootID) {
          Host.rootInstances[instance.rootID] = instance;
          Host.rootComponents[instance.rootID] = instance._internal;
        }
      }
    },
    get: function get(node) {
      return node[KEY];
    },
    remove: function remove(node) {
      var instance = this.get(node);

      if (instance) {
        node[KEY] = null;

        if (instance.rootID) {
          delete Host.rootComponents[instance.rootID];
          delete Host.rootInstances[instance.rootID];
        }
      }
    },
    mount: function mount(element, container, parentInstance) {

      var driver = Host.driver; // Before render callback

      driver.beforeRender && driver.beforeRender(); // Real native root node is body

      if (container == null) {
        container = driver.createBody();
      } // Get the context from the conceptual parent component.


      var parentContext;

      if (parentInstance) {
        var parentInternal = parentInstance._internal;
        parentContext = parentInternal._processChildContext(parentInternal._context);
      } // Update root component


      var prevRootInstance = this.get(container);

      if (prevRootInstance && prevRootInstance.rootID) {
        if (parentContext) {
          // Using _penddingContext to pass new context
          prevRootInstance._internal._penddingContext = parentContext;
        }

        prevRootInstance.update(element);
        return prevRootInstance;
      } // Init root component with empty children


      var renderedComponent = instantiateComponent(createElement(Root));
      var defaultContext = parentContext || {};
      var rootInstance = renderedComponent.mountComponent(container, null, defaultContext);
      this.set(container, rootInstance); // Mount new element through update queue avoid when there is in rendering phase

      rootInstance.update(element); // After render callback

      driver.afterRender && driver.afterRender(rootInstance); // Devtool render new root hook

      Host.hook.Mount._renderNewRootComponent(rootInstance._internal);

      return rootInstance;
    }
  };

  var Hook = {
    ComponentTree: {
      getClosestInstanceFromNode: function getClosestInstanceFromNode(node) {
        return instance.get(node);
      },
      getNodeFromInstance: function getNodeFromInstance(inst) {
        // inst is an internal instance (but could be a composite)
        while (inst._renderedComponent) {
          inst = inst._renderedComponent;
        }

        if (inst) {
          return inst._nativeNode;
        } else {
          return null;
        }
      }
    },
    Mount: {
      _instancesByReactRootID: Host.rootComponents,
      // Stub - React DevTools expects to find this method and replace it
      // with a wrapper in order to observe new root components being added
      _renderNewRootComponent: function _renderNewRootComponent() {}
    },
    Reconciler: {
      // Stubs - React DevTools expects to find these methods and replace them
      // with wrappers in order to observe components being mounted, updated and
      // unmounted
      mountComponent: function mountComponent() {},
      receiveComponent: function receiveComponent() {},
      unmountComponent: function unmountComponent() {}
    },
    // monitor the info of all components
    monitor: null
  };

  /* global __REACT_DEVTOOLS_GLOBAL_HOOK__ */

  if (typeof __REACT_DEVTOOLS_GLOBAL_HOOK__ !== 'undefined' && typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.inject === 'function') {
    __REACT_DEVTOOLS_GLOBAL_HOOK__.inject(Hook);
  }

  var ValueEmitter =
  /*#__PURE__*/
  function () {
    function ValueEmitter(defaultValue) {
      this.handlers = [];
      this.value = defaultValue;
    }

    var _proto = ValueEmitter.prototype;

    _proto.on = function on(handler) {
      this.handlers.push(handler);
    };

    _proto.off = function off(handler) {
      this.handlers = this.handlers.filter(function (h) {
        return h !== handler;
      });
    };

    _proto.emit = function emit() {
      var _this = this;

      this.handlers.forEach(function (handler) {
        return handler(_this.value);
      });
    };

    return ValueEmitter;
  }();

  var uniqueId = 0;
  function createContext(defaultValue) {
    var _Provider$childContex, _Consumer$contextType;

    var contextProp = '__context_' + uniqueId++ + '__';

    var Provider =
    /*#__PURE__*/
    function (_Component) {
      _inheritsLoose(Provider, _Component);

      function Provider() {
        var _this2;

        _this2 = _Component.call(this) || this;
        _this2.emitter = new ValueEmitter(defaultValue);
        return _this2;
      }

      var _proto2 = Provider.prototype;

      _proto2.getChildContext = function getChildContext() {
        var _ref;

        return _ref = {}, _ref[contextProp] = this.emitter, _ref;
      };

      _proto2.componentWillMount = function componentWillMount() {
        if (this.props.value !== undefined) {
          this.emitter.value = this.props.value;
        }
      };

      _proto2.componentWillReceiveProps = function componentWillReceiveProps(nextProps) {
        if (this.props.value !== nextProps.value) {
          this.emitter.value = nextProps.value;
        }
      };

      _proto2.componentDidUpdate = function componentDidUpdate(prevProps) {
        if (this.props.value !== prevProps.value) {
          this.emitter.emit();
        }
      };

      _proto2.render = function render() {
        return this.props.children;
      };

      return Provider;
    }(Component);

    Provider.childContextTypes = (_Provider$childContex = {}, _Provider$childContex[contextProp] = function () {}, _Provider$childContex);
    Provider.contextProp = contextProp;
    Provider.defaultValue = defaultValue;

    var Consumer =
    /*#__PURE__*/
    function (_Component2) {
      _inheritsLoose(Consumer, _Component2);

      function Consumer(props, context) {
        var _this3;

        _this3 = _Component2.call(this, props, context) || this;
        _this3.state = {
          value: _this3.readContext(_this3.context)
        };

        _this3.onUpdate = function (value) {
          return _this3.state.value !== value && _this3.setState({
            value: value
          });
        };

        return _this3;
      }

      var _proto3 = Consumer.prototype;

      _proto3.readContext = function readContext(context) {
        return context[contextProp] ? context[contextProp].value : defaultValue;
      };

      _proto3.componentDidMount = function componentDidMount() {
        if (this.context[contextProp]) {
          this.context[contextProp].on(this.onUpdate);
        }
      };

      _proto3.componentWillReceiveProps = function componentWillReceiveProps(nextProps, nextContext) {
        var newContextValue = this.readContext(nextContext);

        if (this.state.value !== newContextValue) {
          this.setState({
            value: newContextValue
          });
        }
      };

      _proto3.componentWillUnmount = function componentWillUnmount() {
        if (this.context[contextProp]) {
          this.context[contextProp].off(this.onUpdate);
        }
      };

      _proto3.render = function render() {
        var children = this.props.children;
        var consumer = Array.isArray(children) ? children[0] : children;

        if (typeof consumer === 'function') {
          return consumer(this.state.value);
        }
      };

      return Consumer;
    }(Component);

    Consumer.contextTypes = (_Consumer$contextType = {}, _Consumer$contextType[contextProp] = function () {}, _Consumer$contextType);
    return {
      Provider: Provider,
      Consumer: Consumer
    };
  }

  function createRef() {
    return {
      current: null
    };
  }

  function enqueueCallback(internal, callback) {
    if (callback) {
      var callbackQueue = internal._pendingCallbacks || (internal._pendingCallbacks = []);
      callbackQueue.push(callback);
    }
  }

  function enqueueState(internal, partialState) {
    if (partialState) {
      var stateQueue = internal._pendingStateQueue || (internal._pendingStateQueue = []);
      stateQueue.push(partialState);
    }
  }

  function runCallbacks(callbacks, context) {
    if (callbacks) {
      for (var i = 0; i < callbacks.length; i++) {
        callbacks[i].call(context);
      }
    }
  }

  function runUpdate(component) {
    var internal = component._internal;

    if (!internal) {
      return;
    }

    Host.isRendering = true; // If updateComponent happens to enqueue any new updates, we
    // shouldn't execute the callbacks until the next render happens, so
    // stash the callbacks first

    var callbacks = internal._pendingCallbacks;
    internal._pendingCallbacks = null;
    var prevElement = internal._currentElement;
    var prevUnmaskedContext = internal._context;
    var nextUnmaskedContext = internal._penddingContext || prevUnmaskedContext;
    internal._penddingContext = undefined;

    if (internal._pendingStateQueue || internal._pendingForceUpdate) {
      internal.updateComponent(prevElement, prevElement, prevUnmaskedContext, nextUnmaskedContext);
    }

    runCallbacks(callbacks, component);
    Host.isRendering = false;
  }

  function mountOrderComparator(c1, c2) {
    return c2._internal._mountID - c1._internal._mountID;
  }

  function performWork() {
    if (Host.isRendering) {
      return;
    }

    var dirtyComponents;
    var component;

    while (Host.dirtyComponents.length > 0) {
      // before next render, we will flush all the PassiveEffects
      flushPassiveEffects(); // stash the dirtyComponents first

      dirtyComponents = Host.dirtyComponents; // Since reconciling a component higher in the owner hierarchy usually (not
      // always -- see shouldComponentUpdate()) will reconcile children, reconcile
      // them before their children by sorting the array.

      dirtyComponents.sort(mountOrderComparator); // reset Host.dirtyComponents, using to collect dirtyComponent
      // generated by the next render

      Host.dirtyComponents = [];

      while (component = dirtyComponents.pop()) {
        runUpdate(component);
      }
    }
  }

  function scheduleWork(component) {
    var dirtyComponents = Host.dirtyComponents;

    if (dirtyComponents.indexOf(component) < 0) {
      dirtyComponents.push(component);
    }

    performWork();
  }

  function flushPassiveEffects() {
    flushBeforeNextRenderCallbacks();
  }
  var Updater = {
    setState: function setState(component, partialState, callback) {
      var internal = component._internal;

      if (!internal) {
        return;
      }

      !Host.isRendering && flushPassiveEffects();
      enqueueState(internal, partialState);
      enqueueCallback(internal, callback); // pending in componentWillReceiveProps and componentWillMount

      if (!internal._pendingState && internal._renderedComponent) {
        scheduleWork(component);
      }
    },
    forceUpdate: function forceUpdate(component, callback) {
      var internal = component._internal;

      if (!internal) {
        return;
      }

      !Host.isRendering && flushPassiveEffects();
      internal._pendingForceUpdate = true;
      enqueueCallback(internal, callback); // pending in componentWillMount

      if (internal._renderedComponent) {
        scheduleWork(component);
      }
    },
    runCallbacks: runCallbacks
  };

  var setImmediatePolyfill = function setImmediatePolyfill(job) {
    return setTimeout(job, 0);
  }; // 0s


  var scheduleImmediateCallback = typeof setImmediate === 'undefined' ? setImmediatePolyfill : setImmediate;
  var cancelImmediateCallback = typeof clearImmediate === 'undefined' ? clearTimeout : clearImmediate;
  var beforeNextRenderCallbacks = [];
  var beforeNextRenderCallbackId;

  var scheduleBeforeNextRenderCallback = function scheduleBeforeNextRenderCallback(callback) {
    if (beforeNextRenderCallbacks.length === 0) {
      beforeNextRenderCallbackId = scheduleImmediateCallback(commitBeforeNextRenderCallbacks);
    }

    beforeNextRenderCallbacks.push(callback);
  };

  function commitBeforeNextRenderCallbacks() {
    var preCallbacks = beforeNextRenderCallbacks;
    beforeNextRenderCallbacks = [];
    preCallbacks.forEach(function (callback) {
      return callback();
    });
    preCallbacks = null; // exec callback maybe schedule a work

    performWork();
  }

  function flushBeforeNextRenderCallbacks() {
    if (beforeNextRenderCallbacks.length !== 0) {
      cancelImmediateCallback(beforeNextRenderCallbackId);
      commitBeforeNextRenderCallbacks();
    }
  }

  function getCurrentRenderingInstance() {
    var currentInstance = Host.component._instance;

    if (currentInstance) {
      return currentInstance;
    } else {
      throw new Error('Hooks can only be called inside a component.');
    }
  }

  function areInputsEqual(inputs, prevInputs) {
    for (var i = 0; i < inputs.length; i++) {
      var val1 = inputs[i];
      var val2 = prevInputs[i];

      if (val1 === val2 && (val1 !== 0 || 1 / val1 === 1 / val2) || val1 !== val1 && val2 !== val2 // eslint-disable-line no-self-compare
      ) {
          continue;
        }

      return false;
    }

    return true;
  }

  function useState(initialState) {
    var currentInstance = getCurrentRenderingInstance();
    var hookId = currentInstance.getCurrentHookId();
    var hooks = currentInstance.hooks;

    if (!hooks[hookId]) {
      // state lazy initializer
      if (typeof initialState === 'function') {
        initialState = initialState();
      }

      var setState = function setState(newState) {
        var current = hooks[hookId][0];

        if (typeof newState === 'function') {
          newState = newState(current);
        }

        if (newState !== current) {
          // This is a render phase update.  After this render pass, we'll restart
          if (Host.component && Host.component._instance === currentInstance) {
            hooks[hookId][0] = newState;
            currentInstance.isRenderScheduled = true;
          } else {
            !Host.isRendering && flushPassiveEffects();
            hooks[hookId][0] = newState;
            currentInstance.update();
          }
        }
      };

      hooks[hookId] = [initialState, setState];
    }

    return hooks[hookId];
  }
  function useContext(context) {
    var currentInstance = getCurrentRenderingInstance();
    return currentInstance.readContext(context);
  }
  function useEffect(effect, inputs) {
    useEffectImpl(effect, inputs, true);
  }
  function useLayoutEffect(effect, inputs) {
    useEffectImpl(effect, inputs);
  }

  function useEffectImpl(effect, inputs, defered) {
    var currentInstance = getCurrentRenderingInstance();
    var hookId = currentInstance.getCurrentHookId();
    var hooks = currentInstance.hooks;
    inputs = inputs != null ? inputs : [effect];

    if (!hooks[hookId]) {
      var create = function create(immediately) {
        if (!immediately && defered) return scheduleBeforeNextRenderCallback(function () {
          return create(true);
        });
        var current = create.current;

        if (current) {
          // Set this to true to prevent re-entrancy
          var previousIsRendering = Host.isRendering;
          Host.isRendering = true;
          destory.current = current();
          create.current = null;
          Host.isRendering = previousIsRendering;
        }
      };

      var destory = function destory(immediately) {
        if (!immediately && defered) return scheduleBeforeNextRenderCallback(function () {
          return destory(true);
        });
        var current = destory.current;

        if (current) {
          // Set this to true to prevent re-entrancy
          var previousIsRendering = Host.isRendering;
          Host.isRendering = true;
          current();
          destory.current = null;
          Host.isRendering = previousIsRendering;
        }
      };

      create.current = effect;
      currentInstance.hooks[hookId] = {
        create: create,
        destory: destory,
        prevInputs: inputs,
        inputs: inputs
      };
      currentInstance.didMountHandlers.push(create);
      currentInstance.willUnmountHandlers.push(destory);
      currentInstance.didUpdateHandlers.push(function () {
        var _hooks$hookId = hooks[hookId],
            prevInputs = _hooks$hookId.prevInputs,
            inputs = _hooks$hookId.inputs,
            create = _hooks$hookId.create;

        if (prevInputs == null || !areInputsEqual(inputs, prevInputs)) {
          destory();
          create();
        }
      });
    } else {
      var hook = hooks[hookId];
      var _create = hook.create,
          prevInputs = hook.inputs;
      hook.inputs = inputs;
      hook.prevInputs = prevInputs;
      _create.current = effect;
    }
  }

  function useImperativeHandle(ref, create, inputs) {
    var nextInputs = inputs != null ? inputs.concat([ref]) : [ref, create];
    useLayoutEffect(function () {
      if (typeof ref === 'function') {
        ref(create());
        return function () {
          return ref(null);
        };
      } else if (ref != null) {
        ref.current = create();
        return function () {
          ref.current = null;
        };
      }
    }, nextInputs);
  }
  function useRef(initialValue) {
    var currentInstance = getCurrentRenderingInstance();
    var hookId = currentInstance.getCurrentHookId();
    var hooks = currentInstance.hooks;

    if (!hooks[hookId]) {
      hooks[hookId] = {
        current: initialValue
      };
    }

    return hooks[hookId];
  }
  function useCallback(callback, inputs) {
    return useMemo(function () {
      return callback;
    }, inputs);
  }
  function useMemo(create, inputs) {
    var currentInstance = getCurrentRenderingInstance();
    var hookId = currentInstance.getCurrentHookId();
    var hooks = currentInstance.hooks;

    if (!hooks[hookId]) {
      hooks[hookId] = [create(), inputs];
    } else {
      var hook = hooks[hookId];
      var prevInputs = hook[1];

      if (!areInputsEqual(inputs, prevInputs)) {
        hook[0] = create();
      }
    }

    return hooks[hookId][0];
  }
  function useReducer(reducer, initialState, initialAction) {
    var currentInstance = getCurrentRenderingInstance();
    var hookId = currentInstance.getCurrentHookId();
    var hooks = currentInstance.hooks;

    if (!hooks[hookId]) {
      if (initialAction) {
        initialState = reducer(initialState, initialAction);
      }

      var dispatch = function dispatch(action) {
        var hook = hooks[hookId]; // reducer will get in the next render, before that we add all
        // actions to the queue

        var queue = hook[2]; // This is a render phase update.  After this render pass, we'll restart

        if (Host.component && Host.component._instance === currentInstance) {
          queue.push(action);
          currentInstance.isRenderScheduled = true;
        } else {
          !Host.isRendering && flushPassiveEffects();
          queue.push(action);
          currentInstance.update();
        }
      };

      return hooks[hookId] = [initialState, dispatch, []];
    }

    var hook = hooks[hookId];
    var queue = hook[2];
    var next = hook[0];

    for (var i = 0; i < queue.length; i++) {
      next = reducer(next, queue[i]);
    }

    hook[0] = next;
    hook[2] = [];
    return hooks[hookId];
  }

  var hasOwnProperty = Object.prototype.hasOwnProperty;
  /**
   * inlined Object.is polyfill to avoid requiring consumers ship their own
   * https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/is
   */

  function is(x, y) {
    // SameValue algorithm
    if (x === y) {
      // Steps 1-5, 7-10
      // Steps 6.b-6.e: +0 != -0
      return x !== 0 || 1 / x === 1 / y;
    } else {
      // Step 6.a: NaN == NaN
      return x !== x && y !== y;
    }
  }
  /**
   * Performs equality by iterating through keys on an object and returning false
   * when any key has values which are not strictly equal between the arguments.
   * Returns true when the values of all keys are strictly equal.
   */


  function shallowEqual(objA, objB) {
    if (is(objA, objB)) {
      return true;
    }

    if (typeof objA !== 'object' || objA === null || typeof objB !== 'object' || objB === null) {
      return false;
    }

    var keysA = Object.keys(objA);
    var keysB = Object.keys(objB);

    if (keysA.length !== keysB.length) {
      return false;
    } // Test for A's keys different from B.


    for (var i = 0; i < keysA.length; i++) {
      if (!hasOwnProperty.call(objB, keysA[i]) || !is(objA[keysA[i]], objB[keysA[i]])) {
        return false;
      }
    }

    return true;
  }

  function memo(type, compare) {
    if (compare === void 0) {
      compare = shallowEqual;
    }

    // Memo could composed
    if (type.compares) {
      type.compares.push(compare);
    } else {
      type.compares = [compare];
    }

    return type;
  }

  function Fragment(props) {
    return props.children;
  }

  /**
   * Empty Component
   */

  var EmptyComponent =
  /*#__PURE__*/
  function () {
    function EmptyComponent() {
      this._currentElement = null;
    }

    var _proto = EmptyComponent.prototype;

    _proto.mountComponent = function mountComponent(parent, parentInstance, context, childMounter) {
      this._parent = parent;
      this._parentInstance = parentInstance;
      this._context = context;
      var instance = {
        _internal: this
      };
      var nativeNode = this.getNativeNode();

      if (childMounter) {
        childMounter(nativeNode, parent);
      } else {
        Host.driver.appendChild(nativeNode, parent);
      }

      return instance;
    };

    _proto.unmountComponent = function unmountComponent(notRemoveChild) {
      if (this._nativeNode && !notRemoveChild) {
        Host.driver.removeChild(this._nativeNode, this._parent);
      }

      this._nativeNode = null;
      this._parent = null;
      this._parentInstance = null;
      this._context = null;
    };

    _proto.updateComponent = function updateComponent() {// Noop
    };

    _proto.getNativeNode = function getNativeNode() {
      // Weex native node
      if (this._nativeNode == null) {
        this._nativeNode = Host.driver.createEmpty();
      }

      return this._nativeNode;
    };

    return EmptyComponent;
  }();

  /*
   * Ref manager
   */
  var Ref = {
    update: function update(prevElement, nextElement, component) {
      var prevRef = prevElement && prevElement.ref || null;
      var nextRef = nextElement && nextElement.ref || null; // Update refs in owner component

      if (prevRef !== nextRef) {
        // Detach prev RenderedElement's ref
        prevRef != null && this.detach(prevElement._owner, prevRef, component); // Attach next RenderedElement's ref

        nextRef != null && this.attach(nextElement._owner, nextRef, component);
      }
    },
    attach: function attach(ownerComponent, ref, component) {
      if (!ownerComponent) {
        throw new Error('You might be adding a ref to a component that was not created inside a component\'s ' + '`render` method, or you have multiple copies of Rax loaded.');
      }

      var instance = component.getPublicInstance();

      if (typeof ref === 'function') {
        ref(instance);
      } else if (typeof ref === 'object') {
        ref.current = instance;
      } else {
        ownerComponent._instance.refs[ref] = instance;
      }
    },
    detach: function detach(ownerComponent, ref, component) {
      if (typeof ref === 'function') {
        // When the referenced component is unmounted and whenever the ref changes, the old ref will be called with null as an argument.
        ref(null);
      } else {
        // Must match component and ref could detach the ref on owner when A's before ref is B's current ref
        var instance = component.getPublicInstance();

        if (typeof ref === 'object' && ref.current === instance) {
          ref.current = null;
        } else if (ownerComponent._instance.refs[ref] === instance) {
          delete ownerComponent._instance.refs[ref];
        }
      }
    }
  };

  function shouldUpdateComponent(prevElement, nextElement) {
    // TODO: prevElement and nextElement could be array
    var prevEmpty = prevElement === null;
    var nextEmpty = nextElement === null;

    if (prevEmpty || nextEmpty) {
      return prevEmpty === nextEmpty;
    }

    var prevType = typeof prevElement;
    var nextType = typeof nextElement;

    if (prevType === 'string' || prevType === 'number') {
      return nextType === 'string' || nextType === 'number';
    } else {
      return prevType === 'object' && nextType === 'object' && prevElement.type === nextElement.type && prevElement.key === nextElement.key;
    }
  }

  var getElementKeyName = (function (children, element, index) {
    var elementKey = element && element.key;
    var hasKey = typeof elementKey === 'string';
    var defaultName = '.' + index.toString(36);

    if (hasKey) {
      var keyName = '$' + elementKey; // Child keys must be unique.

      var keyUnique = children[keyName] === undefined; // Only the first child will be used when encountered two children with the same key

      if (!keyUnique) console.warn("Encountered two children with the same key \"" + elementKey + "\".");
      return keyUnique ? keyName : defaultName;
    } else {
      return defaultName;
    }
  });

  var STYLE = 'style';
  var CHILDREN = 'children';
  var TREE = 'tree';
  var EVENT_PREFIX_REGEXP = /^on[A-Z]/;
  /**
   * Native Component
   */

  var NativeComponent =
  /*#__PURE__*/
  function () {
    function NativeComponent(element) {
      this._currentElement = element;
    }

    var _proto = NativeComponent.prototype;

    _proto.mountComponent = function mountComponent(parent, parentInstance, context, childMounter) {
      // Parent native element
      this._parent = parent;
      this._parentInstance = parentInstance;
      this._context = context;
      this._mountID = Host.mountID++;
      var props = this._currentElement.props;
      var type = this._currentElement.type;
      var instance$$1 = {
        _internal: this,
        type: type,
        props: props
      };
      var appendType = props.append || TREE; // Default is node

      this._instance = instance$$1; // Clone a copy for style diff

      this._prevStyleCopy = Object.assign({}, props.style);
      var nativeNode = this.getNativeNode();

      if (appendType !== TREE) {
        if (childMounter) {
          childMounter(nativeNode, parent);
        } else {
          Host.driver.appendChild(nativeNode, parent);
        }
      }

      if (this._currentElement && this._currentElement.ref) {
        Ref.attach(this._currentElement._owner, this._currentElement.ref, this);
      } // Process children


      var children = props.children;

      if (children != null) {
        this.mountChildren(children, context);
      }

      if (appendType === TREE) {
        if (childMounter) {
          childMounter(nativeNode, parent);
        } else {
          Host.driver.appendChild(nativeNode, parent);
        }
      }

      Host.hook.Reconciler.mountComponent(this);
      return instance$$1;
    };

    _proto.mountChildren = function mountChildren(children, context) {
      var _this = this;

      if (!Array.isArray(children)) {
        children = [children];
      }

      var renderedChildren = this._renderedChildren = {};
      var renderedChildrenImage = children.map(function (element, index) {
        var renderedChild = instantiateComponent(element);
        var name = getElementKeyName(renderedChildren, element, index);
        renderedChildren[name] = renderedChild;
        renderedChild._mountIndex = index; // Mount

        var mountImage = renderedChild.mountComponent(_this.getNativeNode(), _this._instance, context, null);
        return mountImage;
      });
      return renderedChildrenImage;
    };

    _proto.unmountChildren = function unmountChildren(notRemoveChild) {
      var renderedChildren = this._renderedChildren;

      if (renderedChildren) {
        for (var name in renderedChildren) {
          var renderedChild = renderedChildren[name];
          renderedChild.unmountComponent(notRemoveChild);
        }

        this._renderedChildren = null;
      }
    };

    _proto.unmountComponent = function unmountComponent(notRemoveChild) {
      if (this._nativeNode) {
        var ref = this._currentElement.ref;

        if (ref) {
          Ref.detach(this._currentElement._owner, ref, this);
        }

        instance.remove(this._nativeNode);

        if (!notRemoveChild) {
          Host.driver.removeChild(this._nativeNode, this._parent);
        }

        Host.driver.removeAllEventListeners(this._nativeNode);
      }

      this.unmountChildren(notRemoveChild);
      Host.hook.Reconciler.unmountComponent(this);
      this._currentElement = null;
      this._nativeNode = null;
      this._parent = null;
      this._parentInstance = null;
      this._context = null;
      this._instance = null;
      this._prevStyleCopy = null;
    };

    _proto.updateComponent = function updateComponent(prevElement, nextElement, prevContext, nextContext) {
      // Replace current element
      this._currentElement = nextElement;
      Ref.update(prevElement, nextElement, this);
      var prevProps = prevElement.props;
      var nextProps = nextElement.props;
      this.updateProperties(prevProps, nextProps);
      this.updateChildren(nextProps.children, nextContext);
      Host.hook.Reconciler.receiveComponent(this);
    };

    _proto.updateProperties = function updateProperties(prevProps, nextProps) {
      var propKey;
      var styleName;
      var styleUpdates;
      var driver = Host.driver;

      for (propKey in prevProps) {
        if (propKey === CHILDREN || nextProps.hasOwnProperty(propKey) || !prevProps.hasOwnProperty(propKey) || prevProps[propKey] == null) {
          continue;
        }

        if (propKey === STYLE) {
          var lastStyle = this._prevStyleCopy;

          for (styleName in lastStyle) {
            if (lastStyle.hasOwnProperty(styleName)) {
              styleUpdates = styleUpdates || {};
              styleUpdates[styleName] = '';
            }
          }

          this._prevStyleCopy = null;
        } else if (EVENT_PREFIX_REGEXP.test(propKey)) {
          if (typeof prevProps[propKey] === 'function') {
            driver.removeEventListener(this.getNativeNode(), propKey.slice(2).toLowerCase(), prevProps[propKey]);
          }
        } else {
          driver.removeAttribute(this.getNativeNode(), propKey, prevProps[propKey]);
        }
      }

      for (propKey in nextProps) {
        var nextProp = nextProps[propKey];
        var prevProp = propKey === STYLE ? this._prevStyleCopy : prevProps != null ? prevProps[propKey] : undefined;

        if (propKey === CHILDREN || !nextProps.hasOwnProperty(propKey) || nextProp === prevProp || nextProp == null && prevProp == null) {
          continue;
        } // Update style


        if (propKey === STYLE) {
          if (nextProp) {
            // Clone property
            nextProp = this._prevStyleCopy = Object.assign({}, nextProp);
          } else {
            this._prevStyleCopy = null;
          }

          if (prevProp != null) {
            // Unset styles on `prevProp` but not on `nextProp`.
            for (styleName in prevProp) {
              if (prevProp.hasOwnProperty(styleName) && (!nextProp || !nextProp.hasOwnProperty(styleName))) {
                styleUpdates = styleUpdates || {};
                styleUpdates[styleName] = '';
              }
            } // Update styles that changed since `prevProp`.


            for (styleName in nextProp) {
              if (nextProp.hasOwnProperty(styleName) && prevProp[styleName] !== nextProp[styleName]) {
                styleUpdates = styleUpdates || {};
                styleUpdates[styleName] = nextProp[styleName];
              }
            }
          } else {
            // Assign next prop when prev style is null
            styleUpdates = nextProp;
          }
        } else if (EVENT_PREFIX_REGEXP.test(propKey)) {
          // Update event binding
          var eventName = propKey.slice(2).toLowerCase();

          if (typeof prevProp === 'function') {
            driver.removeEventListener(this.getNativeNode(), eventName, prevProp, nextProps);
          }

          if (typeof nextProp === 'function') {
            driver.addEventListener(this.getNativeNode(), eventName, nextProp, nextProps);
          }
        } else {
          // Update other property
          var payload = {};
          payload[propKey] = nextProp;

          if (nextProp != null) {
            driver.setAttribute(this.getNativeNode(), propKey, nextProp);
          } else {
            driver.removeAttribute(this.getNativeNode(), propKey, prevProps[propKey]);
          }
        }
      }

      if (styleUpdates) {

        driver.setStyles(this.getNativeNode(), styleUpdates);
      }
    };

    _proto.updateChildren = function updateChildren(nextChildrenElements, context) {
      var _this2 = this;

      // prev rendered children
      var prevChildren = this._renderedChildren;
      var driver = Host.driver;

      if (nextChildrenElements == null && prevChildren == null) {
        return;
      }

      var nextChildren = {};
      var oldNodes = {};

      if (nextChildrenElements != null) {
        if (!Array.isArray(nextChildrenElements)) {
          nextChildrenElements = [nextChildrenElements];
        } // Update next children elements


        for (var index = 0, length = nextChildrenElements.length; index < length; index++) {
          var nextElement = nextChildrenElements[index];
          var name = getElementKeyName(nextChildren, nextElement, index);
          var prevChild = prevChildren && prevChildren[name];
          var prevElement = prevChild && prevChild._currentElement;
          var prevContext = prevChild && prevChild._context;

          if (prevChild != null && shouldUpdateComponent(prevElement, nextElement)) {
            if (prevElement !== nextElement || prevContext !== context) {
              // Pass the same context when updating chidren
              prevChild.updateComponent(prevElement, nextElement, context, context);
            }

            nextChildren[name] = prevChild;
          } else {
            // Unmount the prevChild when nextChild is different element type.
            if (prevChild) {
              var oldNativeNode = prevChild.getNativeNode(); // Delay remove child

              prevChild.unmountComponent(true);
              oldNodes[name] = oldNativeNode;
            } // The child must be instantiated before it's mounted.


            nextChildren[name] = instantiateComponent(nextElement);
          }
        }
      }

      var firstPrevChild;
      var delayRemoveFirstPrevChild; // Unmount children that are no longer present.

      if (prevChildren != null) {
        for (var _name in prevChildren) {
          if (!prevChildren.hasOwnProperty(_name)) {
            continue;
          }

          var _prevChild = prevChildren[_name];
          var shouldRemove = !nextChildren[_name]; // Store old first child ref for append node ahead and maybe delay remove it

          if (!firstPrevChild) {
            firstPrevChild = _prevChild;
            delayRemoveFirstPrevChild = shouldRemove;
          } else if (shouldRemove) {
            _prevChild.unmountComponent();
          }
        }
      }

      if (nextChildren != null) {
        (function () {
          // `nextIndex` will increment for each child in `nextChildren`, but
          // `lastIndex` will be the last index visited in `prevChildren`.
          var lastIndex = 0;
          var nextIndex = 0;
          var lastPlacedNode = null;
          var nextNativeNode = [];

          var _loop = function _loop(_name2) {
            if (!nextChildren.hasOwnProperty(_name2)) {
              return "continue";
            }

            var nextChild = nextChildren[_name2];
            var prevChild = prevChildren && prevChildren[_name2];

            if (prevChild === nextChild) {
              var prevChildNativeNode = prevChild.getNativeNode(); // Convert to array type

              if (!Array.isArray(prevChildNativeNode)) {
                prevChildNativeNode = [prevChildNativeNode];
              } // If the index of `child` is less than `lastIndex`, then it needs to
              // be moved. Otherwise, we do not need to move it because a child will be
              // inserted or moved before `child`.


              if (prevChild._mountIndex < lastIndex) {
                // Get the last child
                if (Array.isArray(lastPlacedNode)) {
                  lastPlacedNode = lastPlacedNode[lastPlacedNode.length - 1];
                }

                for (var _i = prevChildNativeNode.length - 1; _i >= 0; _i--) {
                  driver.insertAfter(prevChildNativeNode[_i], lastPlacedNode);
                }
              }

              nextNativeNode = nextNativeNode.concat(prevChildNativeNode);
              lastIndex = Math.max(prevChild._mountIndex, lastIndex); // Update to the latest mount order

              prevChild._mountIndex = nextIndex;
            } else {
              if (prevChild != null) {
                // Update `lastIndex` before `_mountIndex` gets unset by unmounting.
                lastIndex = Math.max(prevChild._mountIndex, lastIndex);
              }

              var parent = _this2.getNativeNode(); // Fragment extended native component, so if parent is fragment should get this._parent


              if (Array.isArray(parent)) {
                parent = _this2._parent;
              }

              nextChild.mountComponent(parent, _this2._instance, context, function (newChild, parent) {
                // TODO: Rework the duplicate code
                var oldChild = oldNodes[_name2];

                if (!Array.isArray(newChild)) {
                  newChild = [newChild];
                }

                function insertNewChild(newChild) {
                  // Get the last child
                  if (Array.isArray(lastPlacedNode)) {
                    lastPlacedNode = lastPlacedNode[lastPlacedNode.length - 1];
                  }

                  var prevFirstNativeNode;

                  if (firstPrevChild && !lastPlacedNode) {
                    prevFirstNativeNode = firstPrevChild.getNativeNode();

                    if (Array.isArray(prevFirstNativeNode)) {
                      prevFirstNativeNode = prevFirstNativeNode[0];
                    }
                  }

                  if (lastPlacedNode) {
                    // Should reverse order when insert new child after lastPlacedNode:
                    // [lastPlacedNode, *newChild1, *newChild2]
                    for (var _i2 = newChild.length - 1; _i2 >= 0; _i2--) {
                      driver.insertAfter(newChild[_i2], lastPlacedNode);
                    }
                  } else if (prevFirstNativeNode) {
                    // [*newChild1, *newChild2, prevFirstNativeNode]
                    for (var _i3 = 0; _i3 < newChild.length; _i3++) {
                      driver.insertBefore(newChild[_i3], prevFirstNativeNode);
                    }
                  } else {
                    // [*newChild1, *newChild2]
                    for (var _i4 = 0; _i4 < newChild.length; _i4++) {
                      driver.appendChild(newChild[_i4], parent);
                    }
                  }
                }

                if (oldChild) {
                  // The oldChild or newChild all maybe fragment
                  if (!Array.isArray(oldChild)) {
                    oldChild = [oldChild];
                  }

                  if (prevChild._mountIndex < lastIndex) {
                    for (var _i5 = 0; _i5 < oldChild.length; _i5++) {
                      driver.removeChild(oldChild[_i5]);
                    }

                    insertNewChild(newChild);
                  } else {
                    // If newChild count large then oldChild:
                    // [oldChild1, oldChild2] => [newChild1, newChild2, newChild3]
                    var lastNewChild;

                    for (var _i6 = 0; _i6 < newChild.length; _i6++) {
                      var child = newChild[_i6];

                      if (oldChild[_i6]) {
                        driver.replaceChild(child, oldChild[_i6]);
                      } else {
                        driver.insertAfter(child, lastNewChild);
                      }

                      lastNewChild = child;
                    } // If newChild count less then oldChild
                    // [oldChild1, oldChild2, oldChild3] => [newChild1, newChild2]


                    if (newChild.length < oldChild.length) {
                      for (var _i7 = newChild.length; _i7 < oldChild.length; _i7++) {
                        driver.removeChild(oldChild[_i7]);
                      }
                    }
                  }
                } else {
                  // Insert child at a specific index
                  insertNewChild(newChild);
                }

                nextNativeNode = nextNativeNode.concat(newChild);
              }); // Update to the latest mount order

              nextChild._mountIndex = nextIndex;
            }

            nextIndex++;
            lastPlacedNode = nextChild.getNativeNode();
          };

          for (var _name2 in nextChildren) {
            var _ret = _loop(_name2);

            if (_ret === "continue") continue;
          } // Sync update native refs


          if (Array.isArray(_this2._nativeNode)) {
            // Clear all and push the new array
            _this2._nativeNode.splice(0, _this2._nativeNode.length);

            for (var i = 0; i < nextNativeNode.length; i++) {
              _this2._nativeNode.push(nextNativeNode[i]);
            }
          }
        })();
      }

      if (delayRemoveFirstPrevChild) {
        firstPrevChild.unmountComponent();
      }

      this._renderedChildren = nextChildren;
    };

    _proto.getNativeNode = function getNativeNode() {
      if (this._nativeNode == null) {
        this._nativeNode = Host.driver.createElement(this._instance);
        instance.set(this._nativeNode, this._instance);
      }

      return this._nativeNode;
    };

    _proto.getPublicInstance = function getPublicInstance() {
      return this.getNativeNode();
    };

    _proto.getName = function getName() {
      return this._currentElement.type;
    };

    return NativeComponent;
  }();

  /**
   * Text Component
   */

  var TextComponent =
  /*#__PURE__*/
  function () {
    function TextComponent(element) {
      this._currentElement = element;
      this._stringText = String(element);
    }

    var _proto = TextComponent.prototype;

    _proto.mountComponent = function mountComponent(parent, parentInstance, context, childMounter) {
      this._parent = parent;
      this._parentInstance = parentInstance;
      this._context = context;
      this._mountID = Host.mountID++; // Weex dom operation

      var nativeNode = this.getNativeNode();

      if (childMounter) {
        childMounter(nativeNode, parent);
      } else {
        Host.driver.appendChild(nativeNode, parent);
      }

      var instance = {
        _internal: this
      };
      Host.hook.Reconciler.mountComponent(this);
      return instance;
    };

    _proto.unmountComponent = function unmountComponent(notRemoveChild) {
      if (this._nativeNode && !notRemoveChild) {
        Host.driver.removeChild(this._nativeNode, this._parent);
      }

      Host.hook.Reconciler.unmountComponent(this);
      this._currentElement = null;
      this._nativeNode = null;
      this._parent = null;
      this._parentInstance = null;
      this._context = null;
      this._stringText = null;
    };

    _proto.updateComponent = function updateComponent(prevElement, nextElement, context) {
      // If some text do noting
      if (prevElement !== nextElement) {
        // Replace current element
        this._currentElement = nextElement; // Devtool read the latest stringText value

        this._stringText = String(nextElement);
        Host.driver.updateText(this.getNativeNode(), this._stringText);
        Host.hook.Reconciler.receiveComponent(this);
      }
    };

    _proto.getNativeNode = function getNativeNode() {
      if (this._nativeNode == null) {
        this._nativeNode = Host.driver.createText(this._stringText);
      }

      return this._nativeNode;
    };

    return TextComponent;
  }();

  var RE_RENDER_LIMIT = 24;
  /**
   * Functional Reactive Component Class Wrapper
   */

  var ReactiveComponent =
  /*#__PURE__*/
  function (_Component) {
    _inheritsLoose(ReactiveComponent, _Component);

    function ReactiveComponent(pureRender, ref) {
      var _this;

      _this = _Component.call(this) || this; // A pure function

      _this.pureRender = pureRender;
      _this.hooksIndex = 0;
      _this.hooks = {};
      _this.didMountHandlers = [];
      _this.didUpdateHandlers = [];
      _this.willUnmountHandlers = [];
      _this.isRenderScheduled = false;
      _this.numberOfReRenders = 0;

      if (pureRender.forwardRef) {
        _this.prevForwardRef = _this.forwardRef = ref;
      }

      var compares = pureRender.compares;

      if (compares) {
        _this.shouldComponentUpdate = function (nextProps) {
          // Process composed compare
          var arePropsEqual = true; // Compare push in and pop out

          for (var i = compares.length - 1; i > -1; i--) {
            if (arePropsEqual = compares[i](_this.props, nextProps)) {
              break;
            }
          }

          return !arePropsEqual || _this.prevForwardRef !== _this.forwardRef;
        };
      }

      return _this;
    }

    var _proto = ReactiveComponent.prototype;

    _proto.getCurrentHookId = function getCurrentHookId() {
      return ++this.hooksIndex;
    };

    _proto.readContext = function readContext(context) {
      var _this2 = this;

      var Provider = context.Provider;
      var unmaskContext = this._internal._context;
      var contextProp = Provider.contextProp;
      var contextEmitter = unmaskContext[contextProp];

      if (contextEmitter) {
        var mountId = this._internal._mountID;

        if (!contextEmitter[mountId]) {
          // One context one updater bind
          contextEmitter[mountId] = {};

          var contextUpdater = function contextUpdater(newContext) {
            if (newContext !== contextEmitter[mountId].renderedContext) {
              _this2.update();
            }
          };

          contextEmitter.on(contextUpdater);
          this.willUnmountHandlers.push(function () {
            delete contextEmitter[mountId];
            contextEmitter.off(contextUpdater);
          });
        }

        return contextEmitter[mountId].renderedContext = contextEmitter.value;
      }

      return Provider.defaultValue;
    };

    _proto.componentDidMount = function componentDidMount() {
      this.didMountHandlers.forEach(function (handler) {
        return handler();
      });
    };

    _proto.componentDidUpdate = function componentDidUpdate() {
      this.didUpdateHandlers.forEach(function (handler) {
        return handler();
      });
    };

    _proto.componentWillUnmount = function componentWillUnmount() {
      this.willUnmountHandlers.forEach(function (handler) {
        return handler();
      });
    };

    _proto.update = function update() {
      this.forceUpdate();
    };

    _proto.render = function render() {

      this.hooksIndex = 0;
      this.numberOfReRenders = 0;
      this.isRenderScheduled = false;
      var children = this.pureRender(this.props, this.forwardRef ? this.forwardRef : this.context);

      while (this.isRenderScheduled) {
        this.numberOfReRenders++;

        if (this.numberOfReRenders > RE_RENDER_LIMIT) {
          throw new Error('Too many re-renders, the number of renders is limited to prevent an infinite loop.');
        }

        this.hooksIndex = 0;
        this.isRenderScheduled = false;
        children = this.pureRender(this.props, this.forwardRef ? this.forwardRef : this.context);
      }

      return children;
    };

    return ReactiveComponent;
  }(Component);

  function performInSandbox(fn, instance, callback) {
    try {
      return fn();
    } catch (e) {
      if (callback) {
        callback(e);
      } else {
        handleError(instance, e);
      }
    }
  }

  function handleError(instance, error) {
    var boundary;

    while (instance) {
      if (typeof instance.componentDidCatch === 'function') {
        boundary = instance;
        break;
      } else if (instance._internal && instance._internal._parentInstance) {
        instance = instance._internal._parentInstance;
      } else {
        break;
      }
    }

    if (boundary) {
      // should not attempt to recover an unmounting error boundary
      var boundaryInternal = boundary._internal;

      if (boundaryInternal) {
        var callbackQueue = boundaryInternal._pendingCallbacks || (boundaryInternal._pendingCallbacks = []);
        callbackQueue.push(function () {
          return boundary.componentDidCatch(error);
        });
      }
    } else {
      // Do not break when error happens
      setTimeout(function () {
        throw error;
      }, 0);
    }
  }
  /**
   * Composite Component
   */


  var CompositeComponent =
  /*#__PURE__*/
  function () {
    function CompositeComponent(element) {
      this._currentElement = element;
    }

    var _proto = CompositeComponent.prototype;

    _proto.getName = function getName() {
      var type = this._currentElement.type;
      var constructor = this._instance && this._instance.constructor;
      return type.displayName || constructor && constructor.displayName || type.name || constructor && constructor.name || null;
    };

    _proto.mountComponent = function mountComponent(parent, parentInstance, context, childMounter) {

      this._parent = parent;
      this._parentInstance = parentInstance;
      this._context = context;
      this._mountID = Host.mountID++;
      this._updateCount = 0;

      var currentElement = this._currentElement;
      var Component = currentElement.type;
      var ref = currentElement.ref;
      var publicProps = currentElement.props;
      var isClass = Component.prototype;
      var isComponentClass = isClass && Component.prototype.isComponentClass; // Class stateless component without state but have lifecycles

      var isStatelessClass = isClass && Component.prototype.render; // Context process

      var publicContext = this._processContext(context); // Initialize the public class


      var instance;
      var renderedElement;

      try {
        if (isComponentClass || isStatelessClass) {
          // Component instance
          instance = new Component(publicProps, publicContext, Updater);
        } else if (typeof Component === 'function') {
          // Functional reactive component with hooks
          instance = new ReactiveComponent(Component, ref);
        } else {
          throw new Error("Invalid component type: " + Component + ". (current: " + (typeof Component === 'object' && Object.keys(Component) || typeof Component) + ")");
        }
      } catch (e) {
        handleError(parentInstance, e);
        return instance;
      } // These should be set up in the constructor, but as a convenience for
      // simpler class abstractions, we set them up after the fact.


      instance.props = publicProps;
      instance.context = publicContext;
      instance.refs = {}; // Inject the updater into instance

      instance.updater = Updater;
      instance._internal = this;
      this._instance = instance; // Init state, must be set to an object or null

      var initialState = instance.state;

      if (initialState === undefined) {
        // TODO clone the state?
        instance.state = initialState = null;
      }

      var error = null;

      var errorCallback = function errorCallback(e) {
        error = e;
      };

      if (instance.componentWillMount) {
        performInSandbox(function () {
          {
            instance.componentWillMount();
          }
        }, instance, errorCallback);
      }

      if (renderedElement == null) {
        Host.component = this; // Process pending state when call setState in componentWillMount

        instance.state = this._processPendingState(publicProps, publicContext);
        performInSandbox(function () {
          {
            renderedElement = instance.render();
          }
        }, instance, errorCallback);
        Host.component = null;
      }

      this._renderedComponent = instantiateComponent(renderedElement);

      this._renderedComponent.mountComponent(this._parent, instance, this._processChildContext(context), childMounter);

      if (error) {
        handleError(instance, error);
      }

      if (!this._currentElement.type.forwardRef && ref) {
        Ref.attach(currentElement._owner, ref, this);
      }

      if (instance.componentDidMount) {
        performInSandbox(function () {
          {
            instance.componentDidMount();
          }
        }, instance);
      } // Trigger setState callback in componentWillMount or boundary callback after rendered


      var callbacks = this._pendingCallbacks;

      if (callbacks) {
        this._pendingCallbacks = null;
        Updater.runCallbacks(callbacks, instance);
      }

      Host.hook.Reconciler.mountComponent(this);

      return instance;
    };

    _proto.unmountComponent = function unmountComponent(notRemoveChild) {
      var instance = this._instance;

      if (!instance) {
        return;
      }

      if (instance.componentWillUnmount) {
        performInSandbox(function () {
          instance.componentWillUnmount();
        }, instance);
      }

      Host.hook.Reconciler.unmountComponent(this);
      instance._internal = null;

      if (this._renderedComponent != null) {
        var ref = this._currentElement.ref;

        if (!this._currentElement.type.forwardRef && ref) {
          Ref.detach(this._currentElement._owner, ref, this);
        }

        this._renderedComponent.unmountComponent(notRemoveChild);

        this._renderedComponent = null;
        this._instance = null;
      }

      this._currentElement = null;
      this._parentInstance = null; // Reset pending fields
      // Even if this component is scheduled for another update in ReactUpdates,
      // it would still be ignored because these fields are reset.

      this._pendingStateQueue = null;
      this._pendingForceUpdate = false; // These fields do not really need to be reset since this object is no
      // longer accessible.

      this._context = null;
    }
    /**
     * Filters the context object to only contain keys specified in
     * `contextTypes`
     */
    ;

    _proto._processContext = function _processContext(context) {
      var Component = this._currentElement.type;
      var contextTypes = Component.contextTypes;

      if (!contextTypes) {
        return {};
      }

      var maskedContext = {};

      for (var contextName in contextTypes) {
        maskedContext[contextName] = context[contextName];
      }

      return maskedContext;
    };

    _proto._processChildContext = function _processChildContext(currentContext) {
      var instance = this._instance;
      var childContext = instance.getChildContext && instance.getChildContext();

      if (childContext) {
        return Object.assign({}, currentContext, childContext);
      }

      return currentContext;
    };

    _proto._processPendingState = function _processPendingState(props, context) {
      var instance = this._instance;
      var queue = this._pendingStateQueue;

      if (!queue) {
        return instance.state;
      } // Reset pending queue


      this._pendingStateQueue = null;
      var nextState = Object.assign({}, instance.state);

      for (var i = 0; i < queue.length; i++) {
        var partial = queue[i];
        Object.assign(nextState, typeof partial === 'function' ? partial.call(instance, nextState, props, context) : partial);
      }

      return nextState;
    };

    _proto.updateComponent = function updateComponent(prevElement, nextElement, prevUnmaskedContext, nextUnmaskedContext) {
      var instance = this._instance;

      if (!instance) {
        console.error("Update component '" + this.getName() + "' that has already been unmounted (or failed to mount).");
      }

      var willReceive = false;
      var nextContext;
      var nextProps; // Determine if the context has changed or not

      if (this._context === nextUnmaskedContext) {
        nextContext = instance.context;
      } else {
        nextContext = this._processContext(nextUnmaskedContext);
        willReceive = true;
      } // Distinguish between a props update versus a simple state update


      if (prevElement === nextElement) {
        // Skip checking prop types again -- we don't read component.props to avoid
        // warning for DOM component props in this upgrade
        nextProps = nextElement.props;
      } else {
        nextProps = nextElement.props;
        willReceive = true;
      }

      var hasReceived = willReceive && instance.componentWillReceiveProps;

      if (hasReceived) {
        // Calling this.setState() within componentWillReceiveProps will not trigger an additional render.
        this._pendingState = true;
        performInSandbox(function () {
          instance.componentWillReceiveProps(nextProps, nextContext);
        }, instance);
        this._pendingState = false;
      } // Update refs


      if (this._currentElement.type.forwardRef) {
        instance.prevForwardRef = prevElement.ref;
        instance.forwardRef = nextElement.ref;
      } else {
        Ref.update(prevElement, nextElement, this);
      } // Shoud update always default


      var shouldUpdate = true;
      var prevProps = instance.props;
      var prevState = instance.state; // TODO: could delay execution processPendingState

      var nextState = this._processPendingState(nextProps, nextContext); // ShouldComponentUpdate is not called when forceUpdate is used


      if (!this._pendingForceUpdate) {
        if (instance.shouldComponentUpdate) {
          shouldUpdate = performInSandbox(function () {
            return instance.shouldComponentUpdate(nextProps, nextState, nextContext);
          }, instance);
        } else if (instance.isPureComponentClass) {
          shouldUpdate = !shallowEqual(prevProps, nextProps) || !shallowEqual(prevState, nextState);
        }
      }

      if (shouldUpdate) {
        this._pendingForceUpdate = false; // Will set `this.props`, `this.state` and `this.context`.

        var prevContext = instance.context; // Cannot use this.setState() in componentWillUpdate.
        // If need to update state in response to a prop change, use componentWillReceiveProps instead.

        performInSandbox(function () {
          if (instance.componentWillUpdate) {
            instance.componentWillUpdate(nextProps, nextState, nextContext);
          }
        }, instance); // Replace with next

        this._currentElement = nextElement;
        this._context = nextUnmaskedContext;
        instance.props = nextProps;
        instance.state = nextState;
        instance.context = nextContext;

        this._updateRenderedComponent(nextUnmaskedContext);

        performInSandbox(function () {
          if (instance.componentDidUpdate) {
            instance.componentDidUpdate(prevProps, prevState, prevContext);
          }
        }, instance);
        this._updateCount++;
      } else {
        // If it's determined that a component should not update, we still want
        // to set props and state but we shortcut the rest of the update.
        this._currentElement = nextElement;
        this._context = nextUnmaskedContext;
        instance.props = nextProps;
        instance.state = nextState;
        instance.context = nextContext;
      } // Flush setState callbacks set in componentWillReceiveProps or boundary callback


      var callbacks = this._pendingCallbacks;

      if (callbacks) {
        this._pendingCallbacks = null;
        Updater.runCallbacks(callbacks, instance);
      }

      Host.hook.Reconciler.receiveComponent(this);
    }
    /**
     * Call the component's `render` method and update the DOM accordingly.
     */
    ;

    _proto._updateRenderedComponent = function _updateRenderedComponent(context) {

      var prevRenderedComponent = this._renderedComponent;
      var prevRenderedElement = prevRenderedComponent._currentElement;
      var instance = this._instance;
      var nextRenderedElement;
      Host.component = this;
      performInSandbox(function () {
        {
          nextRenderedElement = instance.render();
        }
      }, instance);
      Host.component = null;

      if (shouldUpdateComponent(prevRenderedElement, nextRenderedElement)) {
        var prevRenderedUnmaskedContext = prevRenderedComponent._context;

        var nextRenderedUnmaskedContext = this._processChildContext(context);

        if (prevRenderedElement !== nextRenderedElement || prevRenderedUnmaskedContext !== nextRenderedUnmaskedContext) {
          prevRenderedComponent.updateComponent(prevRenderedElement, nextRenderedElement, prevRenderedUnmaskedContext, nextRenderedUnmaskedContext);
        }
      } else {
        var oldChild = prevRenderedComponent.getNativeNode();
        prevRenderedComponent.unmountComponent(true);
        this._renderedComponent = instantiateComponent(nextRenderedElement);

        this._renderedComponent.mountComponent(this._parent, instance, this._processChildContext(context), function (newChild, parent) {
          // TODO: Duplicate code in native component file
          if (!Array.isArray(newChild)) {
            newChild = [newChild];
          } // oldChild or newChild all maybe fragment


          if (!Array.isArray(oldChild)) {
            oldChild = [oldChild];
          } // If newChild count large then oldChild


          var lastNewChild;

          for (var i = 0; i < newChild.length; i++) {
            var child = newChild[i];

            if (oldChild[i]) {
              Host.driver.replaceChild(child, oldChild[i]);
            } else if (lastNewChild) {
              Host.driver.insertAfter(child, lastNewChild);
            } else {
              Host.driver.appendChild(child, parent);
            }

            lastNewChild = child;
          } // If newChild count less then oldChild


          if (newChild.length < oldChild.length) {
            for (var _i = newChild.length; _i < oldChild.length; _i++) {
              Host.driver.removeChild(oldChild[_i]);
            }
          }
        });
      }
    };

    _proto.getNativeNode = function getNativeNode() {
      var renderedComponent = this._renderedComponent;

      if (renderedComponent) {
        return renderedComponent.getNativeNode();
      }
    };

    _proto.getPublicInstance = function getPublicInstance() {
      var instance = this._instance; // The Stateless components cannot be given refs

      if (instance instanceof ReactiveComponent) {
        return null;
      }

      return instance;
    };

    return CompositeComponent;
  }();

  /**
   * Fragment Component
   */

  var FragmentComponent =
  /*#__PURE__*/
  function (_NativeComponent) {
    _inheritsLoose(FragmentComponent, _NativeComponent);

    function FragmentComponent(element) {
      return _NativeComponent.call(this, element) || this;
    }

    var _proto = FragmentComponent.prototype;

    _proto.mountComponent = function mountComponent(parent, parentInstance, context, childMounter) {
      // Parent native element
      this._parent = parent;
      this._parentInstance = parentInstance;
      this._context = context;
      this._mountID = Host.mountID++;
      var instance$$1 = {
        _internal: this
      };
      this._instance = instance$$1;
      var fragment = this.getNativeNode();
      var children = this._currentElement; // Process children

      this.mountChildren(children, context);

      if (childMounter) {
        childMounter(fragment, parent);
      } else {
        for (var i = 0; i < fragment.length; i++) {
          var child = fragment[i];
          Host.driver.appendChild(child, parent);
        }
      }

      return instance$$1;
    };

    _proto.mountChildren = function mountChildren(children, context) {
      var _this = this;

      var renderedChildren = this._renderedChildren = {};
      var fragment = this.getNativeNode();
      var renderedChildrenImage = children.map(function (element, index) {
        var renderedChild = instantiateComponent(element);
        var name = getElementKeyName(renderedChildren, element, index);
        renderedChildren[name] = renderedChild;
        renderedChild._mountIndex = index; // Mount

        var mountImage = renderedChild.mountComponent(_this._parent, _this._instance, context, function (nativeNode) {
          if (Array.isArray(nativeNode)) {
            for (var i = 0; i < nativeNode.length; i++) {
              fragment.push(nativeNode[i]);
            }
          } else {
            fragment.push(nativeNode);
          }
        });
        return mountImage;
      });
      return renderedChildrenImage;
    };

    _proto.unmountComponent = function unmountComponent(notRemoveChild) {
      if (this._nativeNode) {
        instance.remove(this._nativeNode);

        if (!notRemoveChild) {
          for (var i = 0; i < this._nativeNode.length; i++) {
            Host.driver.removeChild(this._nativeNode[i]);
          }
        }
      } // Do not need remove child when their parent is removed


      this.unmountChildren(true);
      this._currentElement = null;
      this._nativeNode = null;
      this._parent = null;
      this._parentInstance = null;
      this._context = null;
      this._instance = null;
    };

    _proto.updateComponent = function updateComponent(prevElement, nextElement, prevContext, nextContext) {
      // Replace current element
      this._currentElement = nextElement;
      this.updateChildren(this._currentElement, nextContext);
    };

    _proto.getNativeNode = function getNativeNode() {
      if (this._nativeNode == null) {
        this._nativeNode = [];
      }

      return this._nativeNode;
    };

    _proto.getPublicInstance = function getPublicInstance() {
      return this.getNativeNode();
    };

    _proto.getName = function getName() {
      return 'fragment';
    };

    return FragmentComponent;
  }(NativeComponent);

  function inject(_ref) {
    var driver = _ref.driver,
        hook = _ref.hook,
        measurer = _ref.measurer;
    // Inject component class
    Host.EmptyComponent = EmptyComponent;
    Host.NativeComponent = NativeComponent;
    Host.TextComponent = TextComponent;
    Host.FragmentComponent = FragmentComponent;
    Host.CompositeComponent = CompositeComponent; // Inject devtool hook

    Host.hook = hook || Hook; // Inject performance measurer

    Host.measurer = measurer; // Inject render driver

    Host.driver = driver || Host.driver;

    if (!Host.driver) {
      throw Error('Driver not found.');
    }
  }

  function render(element, container, options, callback) {
    // Compatible with `render(element, container, callback)`
    if (typeof options === 'function') {
      callback = options;
      options = null;
    }

    options = options || {}; // Init inject

    inject(options);
    var rootComponent = instance.mount(element, container, options.parent);
    var componentInstance = rootComponent.getPublicInstance();

    if (callback) {
      callback.call(componentInstance);
    }

    return componentInstance;
  }

  var version = '1.0.0-beta';

  exports.createElement = createElement;
  exports.createRef = createRef;
  exports.createContext = createContext;
  exports.useState = useState;
  exports.useContext = useContext;
  exports.useEffect = useEffect;
  exports.useLayoutEffect = useLayoutEffect;
  exports.useRef = useRef;
  exports.useCallback = useCallback;
  exports.useMemo = useMemo;
  exports.useReducer = useReducer;
  exports.useImperativeHandle = useImperativeHandle;
  exports.Fragment = Fragment;
  exports.memo = memo;
  exports.render = render;
  exports.version = version;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
//# sourceMappingURL=rax.js.map
